import torch
import torch.nn as nn
from moltx import nets, tokenizers


class AdaMRTokenizerConfig:
    Pretrain = tokenizers.MoltxPretrainConfig(
        token_size=512,
        fmt='smiles',
        spe=True,
        spe_dropout=0.2,
        spe_merges=240
    )

    Generation = tokenizers.MoltxPretrainConfig(
        token_size=512,
        fmt='smiles',
        spe=False,
        spe_dropout=1.0,
        spe_merges=240
    )

    Prediction = tokenizers.MoltxPretrainConfig(
        token_size=512,
        fmt='smiles',
        spe=True,
        spe_dropout=0.0,
        spe_merges=240
    )


class AdaMR(nets.AbsPosEncoderDecoder):
    CONFIG_LARGE = nets.AbsPosEncoderDecoderConfig(
        token_size=512,  # max(spe_merge) = 240
        max_len=512,
        d_model=768,
        nhead=12,
        num_encoder_layers=12,
        num_decoder_layers=12,
        dropout=0.1,
        dtype=torch.bfloat16
    )

    CONFIG_BASE = nets.AbsPosEncoderDecoderConfig(
        token_size=512,  # max(spe_merge) = 240
        max_len=512,
        d_model=768,
        nhead=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dropout=0.1,
        dtype=torch.float32
    )

    def __init__(self, conf: nets.AbsPosEncoderDecoderConfig = CONFIG_LARGE) -> None:
        super().__init__(conf=conf)

    def forward(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        return super().forward_generation(src, tgt)


class AdaMRClassifier(AdaMR):
    def __init__(self, num_classes: int, conf: nets.AbsPosEncoderDecoderConfig) -> None:
        super().__init__(conf=conf)
        d_hidden = conf.d_model // 2
        self.fc = nn.Sequential(
            nn.Dropout(conf.dropout),
            nn.Linear(conf.d_model, d_hidden, dtype=conf.dtype),
            nn.Tanh(),
            nn.Dropout(conf.dropout),
            nn.Linear(d_hidden, num_classes, dtype=conf.dtype)
        )

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')), strict=False)

    def forward(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        feats = super().forward_feature(src, tgt)
        return self.fc(feats)


class AdaMRRegression(AdaMR):
    def __init__(self, conf: nets.AbsPosEncoderDecoderConfig) -> None:
        super().__init__(conf=conf)
        d_hidden = conf.d_model // 2
        self.fc = nn.Sequential(
            nn.Dropout(conf.dropout),
            nn.Linear(conf.d_model, d_hidden, dtype=conf.dtype),
            nn.Tanh(),
            nn.Dropout(conf.dropout),
            nn.Linear(d_hidden, 1, dtype=conf.dtype)
        )

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')), strict=False)

    def forward(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        feats = super().forward_feature(src, tgt)
        return self.fc(feats)


class AdaMRDistGeneration(AdaMR):
    def forward(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        return super().forward_generation(src, tgt)


class AdaMRGoalGeneration(AdaMR):
    def forward(self, goal: torch.Tensor, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        is_batched = src.dim() == 2
        mask = nn.Transformer.generate_square_subsequent_mask(
            tgt.size(-1), device=tgt.device)
        src = self.embedding(src)
        tgt = self.embedding(tgt)
        if is_batched:
            src[:, 0] = src[:, 0] * goal
        else:
            src[0] = src[0] * goal
        out = self.transformer(src, tgt, tgt_mask=mask, tgt_is_causal=True)
        return self.token_output(out)


class AdaMR2(nets.AbsPosEncoderCausal):
    CONFIG_LARGE = nets.AbsPosEncoderCausalConfig(
        token_size=512,  # max(spe_merge) = 240
        max_len=512,
        d_model=768,
        nhead=12,
        num_layers=12,
        dropout=0.1,
        dtype=torch.float32
    )

    CONFIG_BASE = nets.AbsPosEncoderCausalConfig(
        token_size=512,  # max(spe_merge) = 240
        max_len=512,
        d_model=768,
        nhead=8,
        num_layers=6,
        dropout=0.1,
        dtype=torch.float32
    )

    def __init__(self, conf: nets.AbsPosEncoderCausalConfig = CONFIG_LARGE) -> None:
        super().__init__(conf=conf)

    def forward(self, tgt: torch.Tensor) -> torch.Tensor:
        return super().forward_generation(tgt)


class AdaMR2Classifier(AdaMR2):
    def __init__(self, num_classes: int, conf: nets.AbsPosEncoderDecoderConfig) -> None:
        super().__init__(conf=conf)
        d_hidden = conf.d_model // 2
        self.fc = nn.Sequential(
            nn.Dropout(conf.dropout),
            nn.Linear(conf.d_model, d_hidden, dtype=conf.dtype),
            nn.Tanh(),
            nn.Dropout(conf.dropout),
            nn.Linear(d_hidden, num_classes, dtype=conf.dtype)
        )

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')), strict=False)

    def forward(self, tgt: torch.Tensor) -> torch.Tensor:
        feats = super().forward_feature(tgt)
        return self.fc(feats)


class AdaMR2Regression(AdaMR2):
    def __init__(self, conf: nets.AbsPosEncoderDecoderConfig) -> None:
        super().__init__(conf=conf)
        d_hidden = conf.d_model // 2
        self.fc = nn.Sequential(
            nn.Dropout(conf.dropout),
            nn.Linear(conf.d_model, d_hidden, dtype=conf.dtype),
            nn.Tanh(),
            nn.Dropout(conf.dropout),
            nn.Linear(d_hidden, 1, dtype=conf.dtype)
        )

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')), strict=False)

    def forward(self, tgt: torch.Tensor) -> torch.Tensor:
        feats = super().forward_feature(tgt)
        return self.fc(feats)


class AdaMR2DistGeneration(AdaMR2):
    def forward(self, tgt: torch.Tensor) -> torch.Tensor:
        return super().forward_generation(tgt)


class AdaMR2GoalGeneration(AdaMR2):
    def forward(self, goal: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        is_batched = tgt.dim() == 2
        mask = nn.Transformer.generate_square_subsequent_mask(
            tgt.size(-1), device=tgt.device)
        tgt = self.embedding(tgt)
        if is_batched:
            tgt[:, 0] = tgt[:, 0] * goal
        else:
            tgt[0] = tgt[0] * goal
        out = self.transformer(tgt, mask=mask, is_causal=True)
        return self.token_output(out)



import os.path
from moltx import tokenizers as tkz


def test_smi_tkz():

    smi = 'CC[N+](C)(C)Cc1ccccc1Br'

    tok = tkz.SmilesAtomwiseTokenizer()

    tokens = tok(smi)
    assert tokens == ['C', 'C', '[N+]', '(', 'C', ')', '(', 'C', ')',
                      'C', 'c', '1', 'c', 'c', 'c', 'c', 'c', '1', 'Br']


def test_safe_tkz():

    safe = 'N18CC[C@H]CC1.O=C6C#CC8.N67.c17ccc2ncnc4c2c1.N45.c15cccc(Br)c1'

    tok = tkz.SmilesAtomwiseTokenizer()

    tokens = tok(safe)
    assert tokens == ['N', '1', '8', 'C', 'C', '[C@H]', 'C', 'C', '1', '.', 'O', '=', 'C', '6', 'C', '#', 'C', 'C', '8', '.', 'N', '6', '7', '.', 'c', '1', '7',
                      'c', 'c', 'c', '2', 'n', 'c', 'n', 'c', '4', 'c', '2', 'c', '1', '.', 'N', '4', '5', '.', 'c', '1', '5', 'c', 'c', 'c', 'c', '(', 'Br', ')', 'c', '1']


def test_num_tkz():
    tok = tkz.NumericalTokenizer()
    number = '123.659'
    tokens = tok(number)

    assert tokens == ['_1_2_', '_2_1_', '_3_0_',
                      '.', '_6_-1_', '_5_-2_', '_9_-3_']


def test_spe_tkz(datadir):
    smi = "CC[N+](C)(C)Cc1ccccc1Br"
    tok = tkz.SmilesTokenizer(codes_path=os.path.join(datadir, 'spe_smiles.txt'))
    tokens = tok(smi)
    assert tokens == ['CC', '[N+](C)', '(C)C', 'c1ccccc1', 'Br']
    assert tok('C') == ['C']


def test_spe_safe_tkz(datadir):
    smi = "N18CC[C@H]CC1.O=C6C#CC8.N67.c17ccc2ncnc4c2c1.N45.c15cccc(Br)c1"
    tok = tkz.SmilesTokenizer(codes_path=os.path.join(datadir, 'spe_safe.txt'))
    tokens = tok(smi)
    assert tokens == ['N18', 'CC[C@H]', 'CC1', '.O=C', '6C', '#', 'CC8',
                      '.N67', '.c17', 'ccc2', 'nc', 'nc4', 'c2c1', '.N45', '.c15', 'cccc(Br)c1']
    assert tok('C') == ['C']


def test_moltx_tkz():
    tok = tkz.MoltxTokenizer(token_size=128)

    tokens = tok.smi2tokens('<pad>')
    assert tokens == ['<pad>']

    tokens = tok.smi2tokens('c1ccccc1<sep><cls>c1ccccc1')
    assert tokens == ['c', '1', 'c', 'c', 'c', 'c', 'c', '1',
                      '<sep>', '<cls>', 'c', '1', 'c', 'c', 'c', 'c', 'c', '1']

    tokens = tok('BrCl<pad>')
    assert tokens[2] == 0
    assert tok.decode(tokens) == 'BrCl<pad>'

    datadir = '/tmp'
    tok.dump(os.path.join(datadir, 'tks_smiles.json'))
    conf = tkz.MoltxPretrainConfig(token_size=128, spe=False, data_dir=datadir)
    tok2 = tkz.MoltxTokenizer.from_pretrain(conf=conf)
    assert tok._tokens == tok2._tokens
    assert tok._token_idx == tok2._token_idx

    stok = tkz.MoltxTokenizer(token_size=2)
    assert stok._token_size == 2
    assert len(stok._tokens) == 2
    tokens = stok('BrCl')
    assert tokens[0] == 1 # <unk>



import json
import typing
import re
import random
import os
from dataclasses import dataclass


class SmilesAtomwiseTokenizer:
    REGEX = r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"

    def __init__(self, exclusive: typing.Optional[typing.Sequence[str]] = None) -> None:
        """
        Tokenize a SMILES molecule at atom-level:
            (1) 'Br' and 'Cl' are two-character tokens
            (2) Symbols with bracket are considered as tokens

        exclusive: A list of specifical symbols with bracket you want to keep. e.g., ['[C@@H]', '[nH]'].
        Other symbols with bracket will be replaced by '<unk>'. default is `None`.
        """
        self._regex = re.compile(self.REGEX)
        self._exclusive = exclusive

    def __call__(self, smiles: str) -> typing.Sequence[str]:
        tokens = self._regex.findall(smiles)
        if self._exclusive:
            for i, token in enumerate(tokens):
                if token.startswith('[') and token not in self._exclusive:
                    tokens[i] = '<unk>'
        return tokens


class SmilesTokenizer:
    """
    Tokenize SMILES based on the learned SPE tokens.

    codes: output file of `learn_SPE()`

    merges: number of learned SPE tokens you want to use. `-1` means using all of them. `1000` means use the most frequent 1000.

    exclusive_tokens: argument that passes to  `atomwise_tokenizer()`

    dropout: See [BPE-Dropout: Simple and Effective Subword Regularization](https://arxiv.org/abs/1910.13267).
    If `dropout` is set to 0, the segmentation is equivalent to the standard BPE; if `dropout` is set to 1, the segmentation splits words into distinct characters.
    """

    def __init__(self, codes_path: typing.Optional[str] = None, dropout: float = 0.0, merges: int = -1, exclusive_tokens: typing.Optional[typing.Sequence[str]] = None) -> None:
        self.smi_tkz = SmilesAtomwiseTokenizer(exclusive_tokens)
        self.dropout = dropout
        self.bpe_codes = {}
        if codes_path is not None:
            with open(codes_path) as f:
                bpe_codes = [tuple(item.strip().split(' ')) for (
                    n, item) in enumerate(f) if (n < merges or merges == -1)]
            for i, item in enumerate(bpe_codes):
                if len(item) != 2:
                    raise RuntimeError(f"Invalid BPE code at line: {i}")
            self.bpe_codes = dict([(code, i)
                                   for (i, code) in enumerate(bpe_codes)])

    def __call__(self, smiles: str) -> typing.Sequence[str]:
        if len(smiles) == 1:
            return [smiles]
        tokens = self.smi_tkz(smiles)
        if self.dropout >= 0.999999 or not self.bpe_codes:
            return tokens
        while len(tokens) > 1:
            pairs = [(self.bpe_codes[pair], i, pair) for (i, pair) in enumerate(zip(tokens, tokens[1:])) if (
                not self.dropout or random.random() > self.dropout) and pair in self.bpe_codes]
            if not pairs:
                break
            # get first merge operation in list of BPE codes
            bigram = min(pairs)[2]
            positions = [i for (rank, i, pair) in pairs if pair == bigram]
            i = 0
            new_tokens = []
            bigram = ''.join(bigram)
            for j in positions:
                # merges are invalid if they start before current position. This can happen if there are overlapping pairs: (x x x -> xx x)
                if j < i:
                    continue
                # all symbols before merged pair
                new_tokens.extend(tokens[i:j])
                new_tokens.append(bigram)  # merged pair
                i = j + 2  # continue after merged pair
            # add all symbols until end of tokens
            new_tokens.extend(tokens[i:])
            tokens = new_tokens
        return tokens


class NumericalTokenizer:
    REGEX = r"([+-]?\d|\.)"

    def __init__(self):
        self._regex = re.compile(self.REGEX)

    def __call__(self, number: str) -> typing.Sequence[str]:
        digits = self._regex.findall(number)
        try:
            dot = digits.index('.')
        except ValueError:
            dot = len(digits)
        tokens = digits.copy()
        for idx, v in enumerate(digits):
            if idx == dot:
                continue
            p = dot - idx
            if idx < dot:
                p -= 1
            t = f'_{v}_{p}_'
            tokens[idx] = t
        return tokens


@dataclass
class MoltxPretrainConfig:
    token_size: int
    fmt: str = 'smiles'
    data_dir: str = os.path.join(os.path.dirname(__file__), 'data')
    spe: bool = True
    spe_dropout: float = 0.0
    spe_merges: int = -1


class MoltxTokenizer:
    REGEX = re.compile(r"<\w{3}>")
    PAD = "<pad>"
    UNK = "<unk>"
    BOS = "<bos>"
    EOS = "<eos>"
    SEP = "<sep>"
    CLS = "<cls>"
    RESERVED = (PAD, UNK, BOS, EOS, SEP, CLS)

    @classmethod
    def from_pretrain(cls, conf: MoltxPretrainConfig) -> 'MoltxTokenizer':
        datadir = conf.data_dir
        kwargs = {
            'token_size': conf.token_size
        }
        if conf.spe:
            kwargs['spe_codes_path'] = os.path.join(datadir, f'spe_{conf.fmt}.txt')
            kwargs['spe_dropout'] = conf.spe_dropout
            kwargs['spe_merges'] = conf.spe_merges
        tkz = cls(**kwargs, freeze=True)
        tkz.load(os.path.join(datadir, f'tks_{conf.fmt}.json'))
        return tkz

    def __init__(self, token_size: int, freeze: bool = False, spe_codes_path: typing.Optional[str] = None, spe_dropout: float = 0.0, spe_merges: int = -1) -> None:
        self._tokens = []
        self._token_idx = {}
        self._token_size = token_size
        self._freeze = freeze
        self._update_tokens(self.RESERVED)
        spe_kwargs = {}
        if spe_codes_path is not None:
            spe_kwargs['codes_path'] = spe_codes_path
            spe_kwargs['merges'] = spe_merges
            spe_kwargs['dropout'] = spe_dropout
        self._smi_tkz = SmilesTokenizer(**spe_kwargs)

    def _update_tokens(self, tokens: typing.Sequence[str]) -> None:
        if self._freeze:
            return
        for token in tokens:
            if token not in self._token_idx:
                t = len(self._tokens)
                if t >= self._token_size:
                    return
                self._token_idx[token] = t
                self._tokens.append(token)

    def _load_tokens(self, tokens: typing.Sequence[str]) -> None:
        freeze = self._freeze
        self._freeze = False
        self._update_tokens(tokens)
        self._freeze = freeze

    def __getitem__(self, item: typing.Union[int, str]) -> str:
        if isinstance(item, int):
            try:
                return self._tokens[item]
            except IndexError:
                return self.UNK
        return self._token_idx.get(item, self._token_idx[self.UNK])

    def __len__(self) -> int:
        return len(self._tokens)

    def loads(self, tokens_json: str) -> 'MoltxTokenizer':
        tokens = json.loads(tokens_json)['tokens']
        return self._load_tokens(tokens)

    def load(self, path: str) -> 'MoltxTokenizer':
        with open(path, 'r') as f:
            return self.loads(f.read())

    def dumps(self) -> str:
        return json.dumps({
            'tokens': self._tokens
        })

    def dump(self, path: str) -> None:
        with open(path, 'w') as f:
            f.write(self.dumps())

    def smi2tokens(self, smiles: str) -> typing.Sequence[str]:
        tokens = []
        m = self.REGEX.search(smiles)
        pos = 0
        while m is not None:
            start, end = m.span()
            if start > pos:
                tokens.extend(self._smi_tkz(smiles[pos:start]))
            tokens.append(m[0])
            pos = end
            m = self.REGEX.search(smiles, pos=pos)
        if len(smiles) > pos:
            tokens.extend(self._smi_tkz(smiles[pos:]))
        self._update_tokens(tokens)
        return tokens

    def decode(self, token_idxs: typing.Sequence[int]) -> str:
        tokens = [self[idx] for idx in token_idxs]
        return ''.join(tokens)

    def encode(self, smiles: str) -> typing.Sequence[int]:
        tokens = self.smi2tokens(smiles)
        return [self[t] for t in tokens]

    def __call__(self, smiles: str) -> typing.Sequence[int]:
        return self.encode(smiles)



import typing
import torch
from moltx import tokenizers


class Base:
    def __init__(self, tokenizer: tokenizers.MoltxTokenizer, device: torch.device = torch.device('cpu')) -> None:
        self.tokenizer = tokenizer
        self.device = device

    def _tokenize(self, smiles: typing.Sequence[str]) -> torch.Tensor:
        tks_list = [self.tokenizer(smi) for smi in smiles]
        size = max(map(len, tks_list))
        out = [self._tokens2tensor(tks, size).unsqueeze(0) for tks in tks_list]
        return torch.concat(out)

    def _tokens2tensor(self, tokens: typing.Sequence[int], size: int) -> torch.Tensor:
        out = torch.zeros(size, dtype=torch.int)
        if len(tokens) > size:
            raise IndexError('the length of tokens is greater than size!')
        for i, tk in enumerate(tokens):
            out[i] = tk
        return out.to(self.device)


class AdaMR(Base):
    def __call__(self, s1: typing.Sequence[str], s2: typing.Sequence[str]) -> typing.Tuple[torch.Tensor]:
        if len(s1) != len(s2):
            raise RuntimeError("the length of s1 and s2 must be the same!")
        bos = self.tokenizer[self.tokenizer.BOS]
        eos = self.tokenizer[self.tokenizer.EOS]
        src = self._tokenize(s1)
        s2tokens = [self.tokenizer(smi) for smi in s2]
        size = max(map(len, s2tokens)) + 1
        tgt = [self._tokens2tensor([bos] + tks, size).unsqueeze(0) for tks in s2tokens]
        out = [self._tokens2tensor(tks + [eos], size).unsqueeze(0) for tks in s2tokens]
        return src, torch.concat(tgt), torch.concat(out)


class AdaMRClassifier(AdaMR):
    def __call__(self, smiles: typing.Sequence[str], labels: typing.Sequence[int]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(labels):
            raise RuntimeError(
                "the length of smiles and labels must be the same!")
        src = self._tokenize(smiles)
        tgt = self._tokenize(
            [f"{self.tokenizer.BOS}{smi}{self.tokenizer.EOS}" for smi in smiles])
        out = torch.tensor(labels, device=self.device).unsqueeze(-1)
        return src, tgt, out


class AdaMRRegression(AdaMR):
    def __call__(self, smiles: typing.Sequence[str], values: typing.Sequence[float]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(values):
            raise RuntimeError(
                "the length of smiles and values must be the same!")
        src = self._tokenize(smiles)
        tgt = self._tokenize(
            [f"{self.tokenizer.BOS}{smi}{self.tokenizer.EOS}" for smi in smiles])
        out = torch.tensor(values, device=self.device).unsqueeze(-1)
        return src, tgt, out


class AdaMRDistGeneration(AdaMR):
    def __call__(self, smiles: typing.Sequence[str]) -> typing.Tuple[torch.Tensor]:
        head = [self.tokenizer.CLS for _ in range(len(smiles))]
        return super().__call__(head, smiles)


class AdaMRGoalGeneration(AdaMR):
    def __call__(self, smiles: typing.Sequence[str], goals: typing.Sequence[float]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(goals):
            raise RuntimeError(
                "the length of smiles and goals must be the same!")
        head = [self.tokenizer.CLS for _ in range(len(smiles))]
        src, tgt, out = super().__call__(head, smiles)
        goal = torch.tensor(goals, device=self.device).unsqueeze(-1)
        return goal, src, tgt, out


class AdaMR2(Base):
    def __call__(self, s1: typing.Sequence[str], s2: typing.Sequence[str]) -> typing.Tuple[torch.Tensor]:
        if len(s1) != len(s2):
            raise RuntimeError("the length of s1 and s2 must be the same!")
        bos = self.tokenizer[self.tokenizer.BOS]
        eos = self.tokenizer[self.tokenizer.EOS]
        s1tokens = [self.tokenizer(smi) for smi in s1]
        s2tokens = [self.tokenizer(smi) for smi in s2]
        tgt_tokens = [tks1 + [bos] + tks2 for tks1, tks2 in zip(s1tokens, s2tokens)]
        out_tokens = [[0] * len(tks1) + tks2 + [eos] for tks1, tks2 in zip(s1tokens, s2tokens)]
        size = max(map(len, out_tokens))
        tgt = torch.concat([self._tokens2tensor(tks, size).unsqueeze(0) for tks in tgt_tokens])
        out = torch.concat([self._tokens2tensor(tks, size).unsqueeze(0) for tks in out_tokens])
        return tgt, out


class AdaMR2Classifier(AdaMR2):
    def __call__(self, smiles: typing.Sequence[str], labels: typing.Sequence[int]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(labels):
            raise RuntimeError(
                "the length of smiles and labels must be the same!")
        tgt = self._tokenize(
            [f"{self.tokenizer.BOS}{smi}{self.tokenizer.EOS}" for smi in smiles])
        out = torch.tensor(labels, device=self.device).unsqueeze(-1)
        return tgt, out


class AdaMR2Regression(AdaMR2):
    def __call__(self, smiles: typing.Sequence[str], values: typing.Sequence[float]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(values):
            raise RuntimeError(
                "the length of smiles and values must be the same!")
        tgt = self._tokenize(
            [f"{self.tokenizer.BOS}{smi}{self.tokenizer.EOS}" for smi in smiles])
        out = torch.tensor(values, device=self.device).unsqueeze(-1)
        return tgt, out


class AdaMR2DistGeneration(AdaMR2):
    def __call__(self, smiles: typing.Sequence[str]) -> typing.Tuple[torch.Tensor]:
        head = [self.tokenizer.CLS for _ in range(len(smiles))]
        return super().__call__(head, smiles)


class AdaMR2GoalGeneration(AdaMR2):
    def __call__(self, smiles: typing.Sequence[str], goals: typing.Sequence[float]) -> typing.Tuple[torch.Tensor]:
        if len(smiles) != len(goals):
            raise RuntimeError(
                "the length of smiles and goals must be the same!")
        head = [self.tokenizer.CLS for _ in range(len(smiles))]
        tgt, out = super().__call__(head, smiles)
        goal = torch.tensor(goals, device=self.device).unsqueeze(-1)
        return goal, tgt, out



import pytest
from moltx import pipelines, models, nets


@pytest.fixture
def adamr_conf():
    return nets.AbsPosEncoderDecoderConfig(
        token_size=16, max_len=32, d_model=8, nhead=2,
        num_encoder_layers=2, num_decoder_layers=2)


@pytest.fixture
def adamr2_conf():
    return nets.AbsPosEncoderCausalConfig(
        token_size=16, max_len=32, d_model=8, nhead=2,num_layers=2)


def test_AdaMR(tokenizer, adamr_conf):
    pipeline = pipelines.AdaMR(tokenizer, models.AdaMR(adamr_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'smiles' in out and 'probability' in out
    assert isinstance(out['smiles'], str)

    out = pipeline("")
    assert 'smiles' in out and 'probability' in out
    assert isinstance(out['smiles'], str)


def test_AdaMRClassifier(tokenizer, adamr_conf):
    pipeline = pipelines.AdaMRClassifier(
        tokenizer, models.AdaMRClassifier(num_classes=2, conf=adamr_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'label' in out and 'probability' in out
    assert isinstance(out['label'], int) and isinstance(out['probability'], float)
    assert out['label'] in range(2)


def test_AdaMRRegression(tokenizer, adamr_conf):
    pipeline = pipelines.AdaMRRegression(
        tokenizer, models.AdaMRRegression(conf=adamr_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'value' in out
    assert isinstance(out['value'], float)


def test_AdaMRDistGeneration(tokenizer, adamr_conf):
    pipeline = pipelines.AdaMRDistGeneration(
        tokenizer, models.AdaMRDistGeneration(adamr_conf))
    out = pipeline(k=1)
    assert 'smiles' in out and 'probabilities' in out
    assert isinstance(out['smiles'], list) and len(out['smiles']) == 1 and isinstance(out['smiles'][0], str)
    assert isinstance(out['probabilities'], list) and len(
        out['probabilities']) == 1


def test_AdaMRGoalGeneration(tokenizer, adamr_conf):
    pipeline = pipelines.AdaMRGoalGeneration(
        tokenizer, models.AdaMRGoalGeneration(adamr_conf))
    out = pipeline(0.48, k=2)
    assert 'smiles' in out and 'probabilities' in out
    assert isinstance(out['smiles'], list) and len(out['smiles']) == 2 and isinstance(out['smiles'][0], str)
    assert isinstance(out['probabilities'], list) and len(
        out['probabilities']) == 2


def test_AdaMR2(tokenizer, adamr2_conf):
    pipeline = pipelines.AdaMR2(tokenizer, models.AdaMR2(adamr2_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'smiles' in out and 'probability' in out
    assert isinstance(out['smiles'], str)

    out = pipeline("")
    assert 'smiles' in out and 'probability' in out
    assert isinstance(out['smiles'], str)


def test_AdaMR2Classifier(tokenizer, adamr2_conf):
    pipeline = pipelines.AdaMR2Classifier(
        tokenizer, models.AdaMR2Classifier(num_classes=2, conf=adamr2_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'label' in out and 'probability' in out
    assert isinstance(out['label'], int) and isinstance(out['probability'], float)
    assert out['label'] in range(2)


def test_AdaMR2Regression(tokenizer, adamr2_conf):
    pipeline = pipelines.AdaMR2Regression(
        tokenizer, models.AdaMR2Regression(conf=adamr2_conf))
    out = pipeline("CC[N+](C)(C)Br")
    assert 'value' in out
    assert isinstance(out['value'], float)


def test_AdaMR2DistGeneration(tokenizer, adamr2_conf):
    pipeline = pipelines.AdaMR2DistGeneration(
        tokenizer, models.AdaMR2DistGeneration(adamr2_conf))
    out = pipeline(k=1)
    assert 'smiles' in out and 'probabilities' in out
    assert isinstance(out['smiles'], list) and len(out['smiles']) == 1 and isinstance(out['smiles'][0], str)
    assert isinstance(out['probabilities'], list) and len(
        out['probabilities']) == 1


def test_AdaMR2GoalGeneration(tokenizer, adamr2_conf):
    pipeline = pipelines.AdaMR2GoalGeneration(
        tokenizer, models.AdaMR2GoalGeneration(adamr2_conf))
    out = pipeline(0.48, k=2)
    assert 'smiles' in out and 'probabilities' in out
    assert isinstance(out['smiles'], list) and len(out['smiles']) == 2 and isinstance(out['smiles'][0], str)
    assert isinstance(out['probabilities'], list) and len(
        out['probabilities']) == 2



import typing
import torch
import torch.nn as nn
from moltx import tokenizers


class Base:
    def __init__(self, tokenizer: tokenizers.MoltxTokenizer, model: nn.Module, device: torch.device = torch.device('cpu')) -> None:
        self.tokenizer = tokenizer
        model = model.to(device)
        model.eval()
        model.requires_grad_(False)
        self.model = model
        self.device = device

    def _tokenize(self, smiles: str) -> torch.Tensor:
        tks = self.tokenizer(smiles)
        size = len(tks)
        return self._tokens2tensor(tks, size)

    def _tokens2tensor(self, tokens: typing.Sequence[int], size: int) -> torch.Tensor:
        out = torch.zeros(size, dtype=torch.int)
        if len(tokens) > size:
            raise IndexError('the length of tokens is greater than size!')
        for i, tk in enumerate(tokens):
            out[i] = tk
        return out.to(self.device)

    @torch.no_grad()
    def _greedy_search(self, tgt: torch.Tensor, **kwds: torch.Tensor) -> typing.Tuple[typing.Sequence[int], float]:
        maxlen = self.model.conf.max_len
        prefixlen = tgt.size(-1)
        eos = self.tokenizer[self.tokenizer.EOS]
        log_prob = torch.zeros(1, device=self.device)
        for _ in range(maxlen - tgt.size(-1)):
            next_log_prob, next_token = self.model(
                tgt=tgt, **kwds)[-1].log_softmax(-1).max(-1, keepdims=True)  # [token_size] max-> []
            if next_token.item() == eos:
                break
            log_prob += next_log_prob
            tgt = torch.concat((tgt, next_token), dim=-1)
        return self.tokenizer.decode(tgt[prefixlen:].tolist()), log_prob.exp().item()

    @torch.no_grad()
    def _random_sample(self, tgt: torch.Tensor, temperature=1, **kwds: torch.Tensor):
        maxlen = self.model.conf.max_len
        prefixlen = tgt.size(-1)
        eos = self.tokenizer[self.tokenizer.EOS]
        log_prob = torch.zeros(1, device=self.device)
        for _ in range(maxlen - tgt.size(-1)):
            next_log_probs = (self.model(tgt=tgt, **kwds)
                              [-1] / temperature).softmax(-1)  # [token_size]
            rand_num = torch.rand((), device=self.device)
            next_token = (next_log_probs.cumsum(-1) <
                          rand_num).sum(-1, keepdims=True)  # [1]
            if next_token.item() == eos:
                break
            log_prob += next_log_probs[next_token].log()
            tgt = torch.concat((tgt, next_token), dim=-1)
        return self.tokenizer.decode(tgt[prefixlen:].tolist()), log_prob.exp().item()

    @torch.no_grad()
    def _beam_search(self, tgt: torch.Tensor, beam_width: int = 3, **kwds: torch.Tensor):
        # tgt: [seqlen]
        # when beam_width == 1, beam search is equal to greedy search
        maxlen = self.model.conf.max_len
        prefixlen = tgt.size(-1)
        eos = self.tokenizer[self.tokenizer.EOS]
        token_size = self.model.conf.token_size
        smiles = []
        probs = []
        log_probs, next_tokens = self.model(tgt=tgt, **kwds)[-1].log_softmax(-1).topk(k=beam_width, dim=0)  # [beam]
        tgt = torch.concat((tgt.unsqueeze(0).repeat(beam_width, 1), next_tokens.unsqueeze(-1)), dim=-1)
        if 'src' in kwds:
            kwds['src'] = kwds['src'].unsqueeze(0).repeat(beam_width, 1)
        log_probs = log_probs.unsqueeze(-1)
        for _ in range(maxlen - tgt.size(-1)):
            next_log_probs = self.model(
                tgt=tgt, **kwds)[:, -1].log_softmax(-1)  # [beam, token_size]
            next_log_probs = (next_log_probs + log_probs).view(-1, 1)  # [beam * tokensize, 1]
            log_probs, idx = next_log_probs.topk(
                k=beam_width, dim=0)  # [beam, 1]
            tgt_idx = idx.div(token_size, rounding_mode="floor")  # [beam, 1]
            next_tokens = idx - tgt_idx * token_size  # [beam, 1]
            meet_end = (next_tokens.squeeze(1).eq(eos).nonzero()).squeeze(1)
            if meet_end.numel() > 0:
                beam_width -= meet_end.size(0)
                probs.extend(log_probs.index_select(0, meet_end).squeeze(1).exp().tolist())
                end_tgt = tgt.index_select(0, tgt_idx.index_select(0, meet_end).squeeze(1))
                smiles.extend(map(self.tokenizer.decode, end_tgt[:, prefixlen:].tolist()))
                if beam_width == 0:
                    return smiles, probs
            not_end = (next_tokens.squeeze(1).ne(eos).nonzero()).squeeze(1)
            log_probs = log_probs.index_select(0, not_end)
            next_tokens = next_tokens.index_select(0, not_end)
            tgt = tgt.index_select(0, tgt_idx.index_select(0, not_end).squeeze(1))
            tgt = torch.concat((tgt, next_tokens), dim=-1)
            if 'src' in kwds:
                kwds['src'] = kwds['src'].index_select(0, tgt_idx.index_select(0, not_end).squeeze(1))
        probs.extend(log_probs.squeeze(1).exp().tolist())
        smiles.extend(map(self.tokenizer.decode, tgt[:, prefixlen:].tolist()))
        return smiles, probs


class AdaMR(Base):

    def _model_args(self, smiles: str) -> typing.Tuple[torch.Tensor]:
        src = self._tokenize(smiles)
        tgt = self._tokenize(self.tokenizer.BOS)
        return src, tgt

    # gentype: greedy, beam
    def __call__(self, smiles: str = "") -> typing.Mapping:
        src, tgt = self._model_args(smiles)
        if len(smiles) > 0:
            meth = self._do_canonicalize
        else:
            meth = self._do_generate
        smi, prob = meth(src, tgt)
        return {
            'smiles': smi,
            'probability': prob
        }

    def _do_generate(self, src: torch.Tensor, tgt: torch.Tensor) -> typing.Mapping:
        return self._random_sample(src=src, tgt=tgt)

    def _do_canonicalize(self, src: torch.Tensor, tgt: torch.Tensor) -> typing.Mapping:
        smiles, probs = self._beam_search(src=src, tgt=tgt, beam_width=3)
        return smiles[0], probs[0]


class AdaMRClassifier(AdaMR):
    def _model_args(self, smiles: str) -> typing.Tuple[torch.Tensor]:
        src = self._tokenize(smiles)
        tgt = self._tokenize(f"{self.tokenizer.BOS}{smiles}{self.tokenizer.EOS}")
        return src, tgt

    def __call__(self, smiles: str) -> typing.Mapping:
        args = self._model_args(smiles)
        out = self.model(*args)
        prob, label = out.softmax(-1).max(-1)
        return {
            'label': label.item(),
            'probability': prob.item()
        }


class AdaMRRegression(AdaMR):
    def _model_args(self, smiles: str) -> typing.Tuple[torch.Tensor]:
        src = self._tokenize(smiles)
        tgt = self._tokenize(f"{self.tokenizer.BOS}{smiles}{self.tokenizer.EOS}")
        return src, tgt

    def __call__(self, smiles: str) -> typing.Mapping:
        args = self._model_args(smiles)
        out = self.model(*args)
        return {
            'value': out.item()
        }


class AdaMRDistGeneration(AdaMR):
    def _model_args(self) -> typing.Tuple[torch.Tensor]:
        return super()._model_args(self.tokenizer.CLS)

    def __call__(self, k: int = 1) -> typing.Mapping:
        assert k <= 10
        src, tgt = self._model_args()
        smis, probs = [], []
        for _ in range(k):
            smi, prob = self._random_sample(src=src, tgt=tgt)
            smis.append(smi)
            probs.append(prob)
        return {
            'smiles': smis,
            'probabilities': probs
        }


class AdaMRGoalGeneration(AdaMR):
    def _model_args(self) -> typing.Tuple[torch.Tensor]:
        return super()._model_args(self.tokenizer.CLS)

    def __call__(self, goal: float, k: int = 1) -> typing.Mapping:
        assert k <= 10
        src, tgt = self._model_args()
        goal = torch.tensor(goal, device=self.device)
        smis, probs = [], []
        for _ in range(k):
            smi, prob = self._random_sample(goal=goal, src=src, tgt=tgt)
            smis.append(smi)
            probs.append(prob)
        return {
            'smiles': smis,
            'probabilities': probs
        }


class AdaMR2(Base):

    # gentype: greedy, beam
    def __call__(self, smiles: str = "") -> typing.Mapping:
        tgt = self._tokenize(f"{smiles}{self.tokenizer.BOS}")
        if len(smiles) > 0:
            meth = self._do_canonicalize
        else:
            meth = self._do_generate

        smi, prob = meth(tgt)
        return {
            'smiles': smi,
            'probability': prob
        }

    def _do_generate(self, tgt: torch.Tensor) -> typing.Mapping:
        return self._random_sample(tgt=tgt)

    def _do_canonicalize(self, tgt: torch.Tensor) -> typing.Mapping:
        smiles, probs = self._beam_search(tgt=tgt)
        return smiles[0], probs[0]


class AdaMR2Classifier(AdaMR):

    def __call__(self, smiles: str) -> typing.Mapping:
        tgt = self._tokenize(f"{self.tokenizer.BOS}{smiles}{self.tokenizer.EOS}")
        out = self.model(tgt)
        prob, label = out.softmax(-1).max(-1)
        return {
            'label': label.item(),
            'probability': prob.item()
        }


class AdaMR2Regression(AdaMR):

    def __call__(self, smiles: str) -> typing.Mapping:
        tgt = self._tokenize(f"{self.tokenizer.BOS}{smiles}{self.tokenizer.EOS}")
        out = self.model(tgt)
        return {
            'value': out.item()
        }


class AdaMR2DistGeneration(AdaMR):

    def __call__(self, k: int = 1) -> typing.Mapping:
        assert k <= 10
        tgt = self._tokenize(f"{self.tokenizer.CLS}{self.tokenizer.BOS}")
        smis, probs = [], []
        for _ in range(k):
            smi, prob = self._random_sample(tgt=tgt)
            smis.append(smi)
            probs.append(prob)
        return {
            'smiles': smis,
            'probabilities': probs
        }


class AdaMR2GoalGeneration(AdaMR):

    def __call__(self, goal: float, k: int = 1) -> typing.Mapping:
        assert k <= 10
        tgt = self._tokenize(f"{self.tokenizer.CLS}{self.tokenizer.BOS}")
        goal = torch.tensor(goal, device=self.device)
        smis, probs = [], []
        for _ in range(k):
            smi, prob = self._random_sample(goal=goal, tgt=tgt)
            smis.append(smi)
            probs.append(prob)
        return {
            'smiles': smis,
            'probabilities': probs
        }



import pytest
from moltx import datasets, models, tokenizers as tkz


def test_Base(tokenizer):
    ds = datasets.Base(tokenizer)
    smiles = ["<bos>CC[N+](C)(C)Br", "<bos>c1ccccc1"]
    out = ds._tokenize(smiles)
    assert out.shape == (2, 11)
    assert out[0, 0].item() == out[1, 0].item() == ds.tokenizer[ds.tokenizer.BOS]
    assert out[1, -1].item() == 0


def test_AdaMR():
    tokenizer = tkz.MoltxTokenizer.from_pretrain(models.AdaMRTokenizerConfig.Pretrain)
    ds = datasets.AdaMR(tokenizer)
    s1 = ["CC[N+]CCBr", "c1ccccc1"]
    s2 = ["CC[N+]CCBr"]
    with pytest.raises(RuntimeError):
        ds(s1, s2)
    s2 = ["CC[N+](C)(C)Br", ""]
    src, tgt, out = ds(s1, s2)
    assert src.size(0) == 2 and src.size(1) <= 8
    assert tgt.shape == out.shape
    assert tgt[0, 1:].eq(out[0, :-1]).all()

def test_AdaMRClassifier(tokenizer):
    ds = datasets.AdaMRClassifier(tokenizer)
    smiles = ["CC[N+]CCBr", "Cc1ccc1"]
    labels = [1, 2]
    with pytest.raises(RuntimeError):
        ds(smiles, labels[:1])
    src, tgt, out = ds(smiles, labels)
    assert src.shape == (2, 7)
    assert tgt.shape == (2, 9)
    assert out.shape == (2, 1)

def test_AdaMRRegression(tokenizer):
    ds = datasets.AdaMRRegression(tokenizer)
    smiles = ["CC[N+]CCBr", "Cc1ccc1"]
    values = [1.1, 1.2]
    with pytest.raises(RuntimeError):
        ds(smiles, values[:1])
    src, tgt, out = ds(smiles, values)
    assert src.shape == (2, 7)
    assert tgt.shape == (2, 9)
    assert out.shape == (2, 1)


def test_AdaMRDistGeneration(tokenizer):
    ds = datasets.AdaMRDistGeneration(tokenizer)
    smiles = ["CC[N+]CCBr", "c1ccc1"]
    src, tgt, out = ds(smiles)
    assert src.shape == (2, 1)
    assert tgt.shape == (2, 7)
    assert out.shape == (2, 7)


def test_AdaMRGoalGeneration(tokenizer):
    ds = datasets.AdaMRGoalGeneration(tokenizer)
    smiles = ["CC[N+]CCBr", "c1ccc1"]
    goals = [1.1, 1.2]
    goal, src, tgt, out = ds(smiles, goals)
    assert goal.shape == (2, 1)
    assert src.shape == (2, 1)
    assert tgt.shape == (2, 7)
    assert out.shape == (2, 7)

def test_AdaMR2(tokenizer):
    ds = datasets.AdaMR2(tokenizer)
    s1 = ["CC[N+]CCBr", "c1ccccc1"]
    s2 = ["CC[N+]CCBr"]
    with pytest.raises(RuntimeError):
        ds(s1, s2)
    s2 = ["CC[N+](C)(C)Br", ""]
    tgt, out = ds(s1, s2)
    assert tgt.shape == out.shape
    assert tgt.size(0) == 2 and tgt.size(1) == 17
    assert tgt[0, 7:].eq(out[0, 6:-1]).all()
    assert out[0, :6].eq(0).all()

def test_AdaMR2Classifier(tokenizer):
    ds = datasets.AdaMR2Classifier(tokenizer)
    smiles = ["CC[N+]CCBr", "Cc1ccc1"]
    labels = [1, 2]
    with pytest.raises(RuntimeError):
        ds(smiles, labels[:1])
    tgt, out = ds(smiles, labels)
    assert tgt.shape == (2, 9)
    assert out.shape == (2, 1)

def test_AdaMR2Regression(tokenizer):
    ds = datasets.AdaMR2Regression(tokenizer)
    smiles = ["CC[N+]CCBr", "Cc1ccc1"]
    values = [1.1, 1.2]
    with pytest.raises(RuntimeError):
        ds(smiles, values[:1])
    tgt, out = ds(smiles, values)
    assert tgt.shape == (2, 9)
    assert out.shape == (2, 1)


def test_AdaMR2DistGeneration(tokenizer):
    ds = datasets.AdaMR2DistGeneration(tokenizer)
    smiles = ["CC[N+]CCBr", "c1ccc1"]
    tgt, out = ds(smiles)
    assert tgt.shape == (2, 8)
    assert out.shape == (2, 8)
    assert out[0, 0].item() == 0
    assert tgt[0, 2:].eq(out[0, 1:-1]).all()


def test_AdaMR2GoalGeneration(tokenizer):
    ds = datasets.AdaMR2GoalGeneration(tokenizer)
    smiles = ["CC[N+]CCBr", "c1ccc1"]
    goals = [1.1, 1.2]
    goal, tgt, out = ds(smiles, goals)
    assert goal.shape == (2, 1)
    assert tgt.shape == (2, 8)
    assert out.shape == (2, 8)
    assert out[0, 0].item() == 0
    assert tgt[0, 2:].eq(out[0, 1:-1]).all()




import torch
import pytest
from moltx import nets


def test_abspos_embedding():
    emb = nets.AbsPosEmbedding(token_size=64, max_len=10, d_model=32)
    x = torch.randint(0, 64, (2, 5))
    x = emb(x)
    assert x.shape == (2, 5, 32)
    x = torch.randint(0, 64, (5,))
    x = emb(x)
    assert x.shape == (5, 32)
    x = torch.randint(0, 64, (2, 11))
    with pytest.raises(IndexError):
        emb(x)
    x = torch.randint(64, 96, (2, 5))
    with pytest.raises(IndexError):
        emb(x)
    x = torch.tensor([[1, 2, 0, 0, 0], [1, 2, 3, 4, 0]])
    x = emb(x)
    assert x[0, 2:, :].eq(0).all()
    assert x[1, 4:, :].eq(0).all()


def test_AbsPosEncoderDecoder():
    conf = nets.AbsPosEncoderDecoderConfig(
        token_size=16, max_len=32, d_model=8, nhead=2,
        num_encoder_layers=2, num_decoder_layers=2)
    model = nets.AbsPosEncoderDecoder(conf=conf)
    model.eval()
    model.requires_grad_(False)

    batch = torch.randint(1, 16, (2, 32))
    out = model.forward_(batch, batch)
    assert out.shape == (2, 32, 8)
    feat = model.forward_feature(batch, batch)
    assert feat.shape == (2, 8)
    assert feat.eq(out[:, -1]).all()
    gen = model.forward_generation(batch, batch)
    assert gen.shape == (2, 32, 16)

    tokens = torch.randint(1, 16, (32,))
    out = model.forward_(tokens, tokens)
    assert out.shape == (32, 8)
    feat = model.forward_feature(tokens, tokens)
    assert feat.shape == (8,)
    assert feat.eq(out[-1]).all()
    gen = model.forward_generation(tokens, tokens)
    assert gen.shape == (32, 16)

    conf = nets.AbsPosEncoderDecoderConfig(
        token_size=16, max_len=32, d_model=8, nhead=2,
        num_encoder_layers=2, num_decoder_layers=2, dtype=torch.bfloat16)
    model = nets.AbsPosEncoderDecoder(conf=conf)
    model.eval()
    model.requires_grad_(False)
    
    batch = torch.randint(1, 16, (2, 32))
    out = model.forward_(batch, batch)
    assert out.shape == (2, 32, 8)


def test_AbsPosEncoderCausal():
    conf = nets.AbsPosEncoderCausalConfig(
        token_size=16, max_len=32, d_model=8, nhead=2, num_layers=2)
    model = nets.AbsPosEncoderCausal(conf=conf)
    model.eval()
    model.requires_grad_(False)

    batch = torch.randint(1, 16, (2, 32))
    out = model.forward_(batch)
    assert out.shape == (2, 32, 8)
    feat = model.forward_feature(batch)
    assert feat.shape == (2, 8)
    assert feat.eq(out[:, -1]).all()
    gen = model.forward_generation(batch)
    assert gen.shape == (2, 32, 16)

    tokens = torch.randint(1, 16, (32,))
    out = model.forward_(tokens)
    assert out.shape == (32, 8)
    feat = model.forward_feature(tokens)
    assert feat.shape == (8,)
    assert feat.eq(out[-1]).all()
    gen = model.forward_generation(tokens)
    assert gen.shape == (32, 16)

    conf = nets.AbsPosEncoderCausalConfig(
        token_size=16, max_len=32, d_model=8, nhead=2, num_layers=2, dtype=torch.bfloat16)
    model = nets.AbsPosEncoderCausal(conf=conf)
    model.eval()
    model.requires_grad_(False)

    batch = torch.randint(1, 16, (2, 32))
    out = model.forward_(batch)
    assert out.shape == (2, 32, 8)



import torch
import torch.nn as nn
from dataclasses import dataclass


class AbsPosEmbedding(nn.Module):
    def __init__(self, token_size: int, max_len: int, d_model: int, dropout: float = 0.1, dtype: torch.dtype = torch.float32) -> None:
        super().__init__()
        self.token_embedding = nn.Embedding(token_size, d_model, padding_idx=0, dtype=dtype)
        self.pos_embedding = nn.Embedding(max_len + 1, d_model, padding_idx=0, dtype=dtype)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xlen = x.size(-1)
        mask = (x > 0).long()
        position = torch.arange(1, xlen + 1).to(x.device)
        position = position * mask
        x = self.token_embedding(x)
        x += self.pos_embedding(position)
        return self.dropout(x)


@dataclass
class AbsPosEncoderDecoderConfig:
    token_size: int
    max_len: int = 512
    d_model: int = 768
    nhead: int = 8
    num_encoder_layers: int = 6
    num_decoder_layers: int = 6
    dropout: float = 0.1
    dtype: torch.dtype = None


class AbsPosEncoderDecoder(nn.Module):
    def __init__(self, conf: AbsPosEncoderDecoderConfig) -> None:
        super().__init__()
        self.conf = conf
        self.embedding = AbsPosEmbedding(
            conf.token_size, conf.max_len, conf.d_model, conf.dropout, conf.dtype)
        self.transformer = nn.Transformer(
            conf.d_model, conf.nhead, conf.num_encoder_layers, conf.num_decoder_layers, dropout=conf.dropout, activation='gelu', batch_first=True, dtype=conf.dtype)
        self.token_output = nn.Linear(
            conf.d_model, conf.token_size, bias=False, dtype=conf.dtype)

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')))

    def forward_(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        mask = nn.Transformer.generate_square_subsequent_mask(
            tgt.size(-1), device=tgt.device)
        src = self.embedding(src)
        tgt = self.embedding(tgt)
        return self.transformer(src, tgt, tgt_mask=mask, tgt_is_causal=True)

    def forward_feature(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        out = self.forward_(src, tgt)
        indices = (tgt > 0).sum(dim=-1, keepdim=True) - 1
        indices = indices.unsqueeze(-1).repeat(*
                                               [1 for _ in range(tgt.dim())], out.shape[-1])
        return torch.gather(input=out, dim=-2, index=indices).squeeze()

    def forward_generation(self, src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
        out = self.forward_(src, tgt)
        return self.token_output(out)


@dataclass
class AbsPosEncoderCausalConfig:
    token_size: int
    max_len: int = 256
    d_model: int = 768
    nhead: int = 8
    num_layers: int = 12
    dropout: float = 0.1
    dtype: torch.dtype = None


class AbsPosEncoderCausal(nn.Module):
    def __init__(self, conf: AbsPosEncoderCausalConfig) -> None:
        super().__init__()
        self.conf = conf
        self.embedding = AbsPosEmbedding(
            conf.token_size, conf.max_len, conf.d_model, conf.dropout, conf.dtype)
        layer = nn.TransformerEncoderLayer(
            conf.d_model, conf.nhead, dropout=conf.dropout, batch_first=True, activation='gelu', dtype=conf.dtype)
        self.transformer = nn.TransformerEncoder(
            layer, conf.num_layers, norm=nn.LayerNorm(conf.d_model))
        self.token_output = nn.Linear(
            conf.d_model, conf.token_size, bias=False, dtype=conf.dtype)

    def load_ckpt(self, *ckpt_files: str) -> None:
        self.load_state_dict(torch.load(
            ckpt_files[0], map_location=torch.device('cpu')))

    def forward_(self, tgt: torch.Tensor) -> torch.Tensor:
        mask = nn.Transformer.generate_square_subsequent_mask(
            tgt.size(-1), device=tgt.device)
        tgt = self.embedding(tgt)
        return self.transformer(tgt, mask=mask, is_causal=True)

    def forward_feature(self, tgt: torch.Tensor) -> torch.Tensor:
        out = self.forward_(tgt)
        indices = (tgt > 0).sum(dim=-1, keepdim=True) - 1
        indices = indices.unsqueeze(-1).repeat(*
                                               [1 for _ in range(tgt.dim())], out.shape[-1])
        return torch.gather(input=out, dim=-2, index=indices).squeeze()

    def forward_generation(self, tgt: torch.Tensor) -> torch.Tensor:
        out = self.forward_(tgt)
        return self.token_output(out)
